@@ -79,7 +79,15 @@ import time
 ############################################################################################
 def load_test_data(test_data_or_file, project_name, cat_vocab_dict="",
                                                  verbose=0):
+    """
+    Load a CSV file and given a project name, it will load the artifacts in project_name folder.
+    Optionally you can provide the artifacts dictionary as "cat_vocab_dict" in this input.
 
+    Outputs:
+    --------
+    data_batches: a tf.data.Dataset which will be Repeat batched dataset
+    cat_vocab_dict: artifacts dictionary that you can feed to the predict function of model.
+    """
     ### load a small sample of data into a pandas dataframe ##
     if isinstance(test_data_or_file, str):
         test_small = pd.read_csv(test_data_or_file) ### this reads the entire file
@@ -112,25 +120,42 @@ def load_test_data(test_data_or_file, project_name, cat_vocab_dict="",
             print('Unable to load pickle file. Continuing...')
             no_cat_vocab_dict = True
     ####################################################
-    target = cat_vocab_dict['target_variables']
-    usecols = cat_vocab_dict['target_variables']
-    if len(target) == 1:
-        target_name = target[0]
+    if not no_cat_vocab_dict:
+        target = cat_vocab_dict['target_variables']
+        usecols = cat_vocab_dict['target_variables']
+        if len(target) == 1:
+            target_name = target[0]
+        else:
+            target_name = target
     else:
-        target_name = target
+        target = []
+        target_name = ''
+        print('no target variable found since model artifacts dictionary could not be found')
     ### classify variables using the small dataframe ##
     model_options = {}
+
     if no_cat_vocab_dict:
         model_options['DS_LEN'] = 10000  ### Just set some default #######
-        var_df, cat_vocab_dict = classify_features_using_pandas(test_small, target=target_name,
+        ###### Just send in entire dataframe to convert and correct dtypes using this function ##
+        ######   If you don't do this, in some data sets due to mixed types it errors ###
+        ######  Just send in target_name as '' since we want even target to be corrected if it
+        #####    has the wrong data type since tensorflow automatically detects data types.
+        test_small, var_df, cat_vocab_dict = classify_features_using_pandas(test_small, target='',
                                     model_options=model_options, verbose=verbose)
         ##########    Just transfer all the values from var_df to cat_vocab_dict  ##########
         for each_key in var_df:
             cat_vocab_dict[each_key] = var_df[each_key]
         ####################################################################################
     else:
+        ###### Just send in entire dataframe to convert and correct dtypes using this function ##
+        ######   If you don't do this, in some data sets due to mixed types it errors ###
+        ######  Just send in target_name as '' since we want even target to be corrected if it
+        #####    has the wrong data type since tensorflow automatically detects data types.
         model_options['DS_LEN'] = cat_vocab_dict['DS_LEN'] ### you need this to classify features
+        test_small, _, _ = classify_features_using_pandas(test_small, target='',
+                                    model_options=model_options, verbose=verbose)
     ############  Now load the file or dataframe into tf.data.DataSet here #################
+
     preds = list(test_small)
     #batch_size = 64   ## artificially set a size ###
     batch_size = cat_vocab_dict["batch_size"]
@@ -160,7 +185,7 @@ def load_test_data(test_data_or_file, project_name, cat_vocab_dict="",
                                                shuffle=False,
                                                num_parallel_reads=tf.data.experimental.AUTOTUNE)
     else:
-
+        #### This is to load dataframes into datasets ########################
         if test_small.isnull().sum().sum() > 0:
             test_small = fill_missing_values_for_TF2(test_small, cat_vocab_dict)
 
@@ -168,6 +193,7 @@ def load_test_data(test_data_or_file, project_name, cat_vocab_dict="",
         if len(drop_cols) > 0:
             print('    Dropping %s columns from dataset...' %drop_cols)
             test_small.drop(drop_cols, axis=1, inplace=True)
+        #### In some datasets, due to mixed data types in test_small, this next line errors. Beware!!
 
         data_batches = tf.data.Dataset.from_tensor_slices(dict(test_small))
         ### batch it if you are creating it from a dataframe
@@ -176,7 +202,10 @@ def load_test_data(test_data_or_file, project_name, cat_vocab_dict="",
     print('    test data loaded successfully.')
 
     if verbose >= 1:
-        print_one_row_from_tf_dataset(data_batches)
+        try:
+            print_one_row_from_tf_dataset(data_batches)
+        except:
+            pass
     #### These are the input variables for which we are going to create keras.Inputs ###\
     return data_batches, cat_vocab_dict
 #################################################################################################
@@ -201,30 +230,51 @@ def find_batch_size(DS_LEN):
     print('    Batch size selected as %d' %batch_len)
     return batch_len
 ###############################################################################################
+class BalancedSparseCategoricalAccuracy(keras.metrics.SparseCategoricalAccuracy):
+    def __init__(self, name='balanced_sparse_categorical_accuracy', dtype=None):
+        super().__init__(name, dtype=dtype)
+
+    def update_state(self, y_true, y_pred, sample_weight=None):
+        y_flat = y_true
+        if y_true.shape.ndims == y_pred.shape.ndims:
+            y_flat = tf.squeeze(y_flat, axis=[-1])
+        y_true_int = tf.cast(y_flat, tf.int32)
+
+        cls_counts = tf.math.bincount(y_true_int)
+        cls_counts = tf.math.reciprocal_no_nan(tf.cast(cls_counts, self.dtype))
+        weight = tf.gather(cls_counts, y_true_int)
+        return super().update_state(y_true, y_pred, sample_weight=weight)
+#####################################################################################
 def load_model_dict(model_or_model_path, cat_vocab_dict, project_name):
     start_time = time.time()
     if not cat_vocab_dict:
         ### No cat_vocab_dict is given. Hence you must load it from disk ###
-        try:
-            pickle_path = os.path.join(project_name, "cat_vocab_dict.pickle")
-            print('\nLoading cat_vocab_dict file using pickle in %s...will take time...' %pickle_path)
-            cat_vocab_dict = pickle.load(open(pickle_path,"rb"))
-            print('    Loaded pickle file in %s' %pickle_path)
+        print('\nNo model artifacts file given. Loading cat_vocab_dict file using pickle. Will take time...')
+        if isinstance(model_or_model_path, str):
+            if model_or_model_path:
+                try:
+                    pickle_path = os.path.join(model_or_model_path,os.path.join("artifacts", "cat_vocab_dict.pickle"))
+                    cat_vocab_dict = pickle.load(open(pickle_path,"rb"))
+                    print('    Loaded pickle file in %s' %pickle_path)
+                except:
+                    print('Unable to load model and data artifacts cat_vocab_dictionary file. Returning...')
+                    return []
             modeltype = cat_vocab_dict['modeltype']
-        except:
-            print('Unable to load model and data artifacts cat_vocab_dictionary file. Returning...')
-            return []
     else:
         ### cat_vocab_dictionary is given #####
             modeltype = cat_vocab_dict['modeltype']
     ### Check if model is available to be loaded #######
+    print('\nLoading deep_autoviml model from %s folder. This will take time...' %model_or_model_path)
     if isinstance(model_or_model_path, str):
         try:
             if model_or_model_path == "":
                 model_or_model_path = os.path.join(project_name, keras_model_type)
-            print('\nLoading deep_autoviml model from %s folder...' %model_or_model_path)
-            model = tf.keras.models.load_model(os.path.join(model_or_model_path))
-            print('    time taken in mins for loading model = %0.0f' %((time.time()-start_time)/60))
+            else:
+                if modeltype == 'Regression':
+                    model = tf.keras.models.load_model(os.path.join(model_or_model_path))
+                else:
+                    model = tf.keras.models.load_model(os.path.join(model_or_model_path),
+                            custom_objects={'balanced_sparse_categorical_accuracy':BalancedSparseCategoricalAccuracy})
         except Exception as error:
             print('Could not load given model.\nError: %s\n Please check your model path and try again.' %error)
             return
@@ -234,54 +284,131 @@ def load_model_dict(model_or_model_path, cat_vocab_dict, project_name):
     print('Time taken to load saved model = %0.0f seconds' %((time.time()-start_time)))
     return model, cat_vocab_dict
 ###################################################################################################
+##########     THIS IS THE MAIN PREDICT() FUNCTION            #####################################
+###################################################################################################
 def predict(model_or_model_path, project_name, test_dataset,
                     keras_model_type, cat_vocab_dict="", verbose=0):
     start_time2 = time.time()
     model, cat_vocab_dict = load_model_dict(model_or_model_path, cat_vocab_dict, project_name)
     ##### load the test data set here #######
+    if keras_model_type.lower() in ['nlp', 'text']:
+        NLP_VARS = cat_vocab_dict['predictors_in_train']
+    else:
+        NLP_VARS = cat_vocab_dict['nlp_vars']
+    ################################################################
+    @tf.function
+    def process_NLP_features(features):
+        """
+        This is how you combine all your string NLP features into a single new feature.
+        Then you can perform embedding on this combined feature.
+        It takes as input an ordered dict named features and returns the same features format.
+        """
+        return tf.strings.reduce_join([features[i] for i in NLP_VARS],axis=0,
+                keepdims=False, separator=' ', name="combined")
+    ################################################################
+    NLP_COLUMN = "combined_nlp_text"
+    ################################################################
+    @tf.function
+    def combine_nlp_text(features):
+        ##use x to derive additional columns u want. Set the shape as well
+        y = {}
+        y.update(features)
+        y[NLP_COLUMN] = tf.strings.reduce_join([features[i] for i in NLP_VARS],axis=0,
+                keepdims=False, separator=' ')
+        return y
+    ################################################################
     if isinstance(test_dataset, str):
         test_ds, cat_vocab_dict2 = load_test_data(test_dataset, project_name=project_name,
                                 cat_vocab_dict=cat_vocab_dict, verbose=verbose)
+        ### You have to load only the NLP or text variables into dataset. otherwise, it will fail during predict
+        if NLP_VARS:
+            if keras_model_type.lower() in ['nlp', 'text']:
+                test_ds = test_ds.map(process_NLP_features)
+                test_ds = test_ds.unbatch().batch(batch_size)
+                print('    combined NLP or text vars: %s into a single feature successfully' %NLP_VARS)
+            else:
+                test_ds = test_ds.map(combine_nlp_text)
+                print('    combined NLP or text vars: %s into a single feature successfully' %NLP_VARS)
+        else:
+            print('No NLP vars in data set. No preprocessing done.')
         batch_size = cat_vocab_dict2["batch_size"]
         DS_LEN = cat_vocab_dict2["DS_LEN"]
         print("test data size = ",DS_LEN, ', batch_size = ',batch_size)
     elif isinstance(test_dataset, pd.DataFrame) or isinstance(test_dataset, pd.Series):
+        if keras_model_type.lower() in ['nlp', 'text']:
+            #### You must only load the text or nlp columns into the dataset. Otherwise, it will fail during predict.
+            test_dataset = test_dataset[NLP_VARS]
         test_ds, cat_vocab_dict2 = load_test_data(test_dataset, project_name=project_name,
                                 cat_vocab_dict=cat_vocab_dict, verbose=verbose)
-        test_small = test_dataset
         batch_size = cat_vocab_dict2["batch_size"]
         DS_LEN = cat_vocab_dict2["DS_LEN"]
         print("test data size = ",DS_LEN, ', batch_size = ',batch_size)
+        if NLP_VARS:
+            if keras_model_type.lower() in ['nlp', 'text']:
+                test_ds = test_ds.map(process_NLP_features)
+                test_ds = test_ds.unbatch().batch(batch_size)
+                print('    processed NLP and text vars: %s successfully' %NLP_VARS)
+            else:
+                test_ds = test_ds.map(combine_nlp_text)
+                print('    combined NLP or text vars: %s into a single combined_nlp_text successfully' %NLP_VARS)
+        else:
+            print('No NLP vars in data set. No preprocessing done.')
     else:
         ### It must be a tf.data.Dataset hence just load it as is ####
+        if cat_vocab_dict:
+            DS_LEN = cat_vocab_dict["DS_LEN"]
+            batch_size = cat_vocab_dict["batch_size"]
+        else:
+            print('Since artifacts dictionary (cat_vocab_dict) not provided, using 100,000 as default test data size and batch size as 64.')
+            print('    if you want to modify them, send in cat_vocab_dict["DS_LEN"] and cat_vocab_dict["batch_size"]')
+            DS_LEN = 100000
+            batch_size = 64
         test_ds = test_dataset
-        DS_LEN = 100000
-        batch_size = 64
+        if NLP_VARS:
+            if keras_model_type.lower() in ['nlp', 'text']:
+                test_ds = test_ds.map(process_NLP_features)
+                test_ds = test_ds.unbatch().batch(batch_size)
+                print('    processed NLP and text vars: %s successfully' %NLP_VARS)
+            else:
+                test_ds = test_ds.map(combine_nlp_text)
+                print('    combined NLP or text vars: %s into a single combined_nlp_text successfully' %NLP_VARS)
+        else:
+            print('No NLP vars in data set. No preprocessing done.')
         cat_vocab_dict2 = copy.deepcopy(cat_vocab_dict)
-    ##### Now you must convert the boolean to string on the data set here ###################
-    BOOLS = cat_vocab_dict['bools']
+    ##################################################################
+    BOOLS = cat_vocab_dict2['bools']
     #################################################################################
+    @tf.function
     def process_boolean_features(features):
+        """
+        This is how you convert all your boolean features into float variables.
+        The reason you have to do this is because tf.keras does not know how to handle boolean types.
+        It takes as input an ordered dict named features and returns the same in features format.
+        """
         for feature_name in features:
             if feature_name in BOOLS:
                 # Cast boolean feature values to string.
                 features[feature_name] = tf.cast(features[feature_name], tf.dtypes.float32)
         return features
     ##################################################################
-    test_ds = test_ds.map(process_boolean_features)
-    print('Boolean column successfully processed')
+    try:
+        test_ds = test_ds.map(process_boolean_features)
+        print('Boolean column successfully processed')
+    except:
+        print('Error in Boolean column processing. Continuing...')
     ## num_steps is needed to predict on whole dataset once ##
     try:
         num_steps = int(np.ceil(DS_LEN/batch_size))
+        print('Batch size = %s' %batch_size)
     except:
         num_steps = 1
     #########  See if you can predict here if not return the null result #####
     print('    number of steps needed to predict: %d' %num_steps)
     y_test_preds_list = []
     targets = cat_vocab_dict2['target_variables']
-
+    ########  This is where we start predictions on test data set ##############
     try:
-        y_probas = model.predict(test_ds, steps=num_steps)
+        y_probas = model.predict(test_ds, steps=num_steps)[:DS_LEN,:]
         if len(targets) == 1:
             y_test_preds_list.append(y_probas)
         else:
