@@ -91,7 +91,7 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
     Remember this is the most valuable part of this entire library!
     This is one humongous preprocessing step to build everything needed for preprocessing into a keras model!
     But it will break in some cases since we cannot handle every known dataset!
-    It will be good enough for most instances to create a fast keras pipeline + baseline model. 
+    It will be good enough for most instances to create a fast keras pipeline + baseline model.
     You can always fine tune it.
     You can always create your own model and feed it once you have successfully created preprocessing pipeline.
     """
@@ -148,8 +148,8 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
         if num_classes == 2:
             class_size = 1
         else:
-            class_size = num_classes    
-    ###### Now calculate some layer sizes #####    
+            class_size = num_classes
+    ###### Now calculate some layer sizes #####
     data_size = cat_vocab_dict["DS_LEN"]
     data_dim = data_size*len(FEATURE_NAMES)
     dense_layer1, dense_layer2, dense_layer3 = get_hidden_layers(data_dim)
@@ -157,7 +157,7 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
     ###########     F E A T U R E    P R E P R O C E S S I N G   H E R E      #######
     #################################################################################
     nlps = var_df['nlp_vars']
-    keras_options, model_options, num_predicts, output_activation = get_model_defaults(keras_options, 
+    keras_options, model_options, num_predicts, output_activation = get_model_defaults(keras_options,
                                     model_options, targets)
     ##################  NLP Text Features are Proprocessed Here  ################
     nlp_inputs = []
@@ -165,7 +165,7 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
     if len(nlps) > 0:
         print('Starting NLP string column layer preprocessing...')
         nlp_inputs, embedding, nlp_names = preprocessing_nlp(train_ds, model_options,
-                                                var_df, cat_vocab_dict, 
+                                                var_df, cat_vocab_dict,
                                                 keras_model_type, verbose)
         ### we call nlp_outputs as embedding in this section of the program ####
         print('    NLP Preprocessing completed.')
@@ -173,8 +173,9 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
         embedding = []
         print('There are no NLP variables in this dataset for preprocessing...')
     ##################  All other Features are Proprocessed Here  ################
-    fast_models = ['fast', 'fast1','fast2','deep_and_cross','deep and cross',
-                        'deep cross','deep_and_wide','deep and wide','deep wide']
+    fast_models = ['fast','deep_and_wide','deep_wide','wide_deep',
+                    'wide_and_deep','deep wide', 'wide deep', 'fast1',
+                    'deep_and_cross', 'deep_cross', 'deep cross', 'fast2']
     ##############################################################################
     meta_outputs = []
     print('Preprocessing non-NLP layers for %s Keras model...' %keras_model_type)
@@ -182,22 +183,13 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
         ################################################################################
         ############ T H I S   I S   F O R  "A U T O"  M O D E L S    O N L Y  #########
         ################################################################################
-        
-        if len(nlps) > 0:
-            ###### This is to transform NLP embeddings into outputs #############
-            x = layers.Bidirectional(tf.keras.layers.LSTM(128, return_sequences=True))(embedding)
-            x = layers.Bidirectional(tf.keras.layers.LSTM(64))(x)
-            nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(x)
-        else:
-            nlp_outputs = embedding
         if len(lats+lons) > 0:
             print('    starting categorical, float and integer layer preprocessing...')
-            meta_outputs, meta_inputs, meta_names = preprocessing_tabular(train_ds, var_df, 
+            meta_outputs, meta_inputs, meta_names = preprocessing_tabular(train_ds, var_df,
                                                         cat_feat_cross_flag, model_options,
                                                         cat_vocab_dict, keras_model_type, verbose)
             print('    All Non-NLP feature preprocessing for %s completed.' %keras_model_type)
 
-            print('    All NLP feature preprocessing for %s completed.' %keras_model_type)
             ### this is the order in which columns have been trained ###
             final_training_order = nlp_names + meta_names
             ### find their dtypes - remember to use element_spec[0] for train data sets!
@@ -207,7 +199,9 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
                 print('Inferred column names, layers and types (double-check for duplicates and correctness!): \n%s' %col_type_tuples)
             print('    %s model loaded and compiled successfully...' %keras_model_type)
         else:
-            ####### Now combine them into a deep and wide model ######################
+            ####### Now combine them into a deep and wide model ##############################
+            ##  Since we are processing NLPs separately we need to remove them from inputs ###
+            FEATURE_NAMES = left_subtract(FEATURE_NAMES, nlps)
             dropout_rate = 0.1
             hidden_units = [dense_layer2, dense_layer3]
             inputs = create_fast_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS)
@@ -220,17 +214,19 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
             meta_inputs = list(inputs.values()) ### convert input layers to a list
             #### If there are NLP vars in dataset, you must combine the nlp_outputs ##
             if len(nlps) > 0:
-                merged = [wide, deep, nlp_outputs]
+                merged = [wide, deep, embedding]
             else:
                 merged = [wide, deep]
             # = layers.Bidirectional(layers.LSTM(dense_layer1, dropout=0.3, recurrent_dropout=0.3,
             #                                    return_sequences=False, batch_size=batch_size,
             #                                    kernel_regularizer=regularizers.l2(0.01)))(x)
             print('    %s model loaded and compiled successfully...' %keras_model_type)
-            return nlp_inputs, meta_inputs, merged, nlp_outputs
+            return nlp_inputs, meta_inputs, merged, embedding
+    else:
+        meta_inputs = []
     ##### You need to send in the ouput from embedding layer to this sequence of layers ####
     nlp_outputs = []
-    if len(nlps) > 0:
+    if not isinstance(embedding, list):
         if keras_model_type.lower() in ['bert','nlp','text', 'use']:
             ###### This is where you define the NLP Embedded Layers ########
             #x = layers.Dense(64, activation='relu')(embedding)
@@ -256,20 +252,18 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
 
         elif keras_model_type.lower() in ['cnn1']:
             # Conv1D + global max pooling
-            x = Conv1D(dense_layer1, 14, name='cnn_dense1', padding="same", 
+            x = Conv1D(dense_layer1, 14, name='cnn_dense1', padding="same",
                                     activation="relu", strides=3)(embedding)
             x = GlobalMaxPooling1D()(x)
             nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(x)
         elif keras_model_type.lower() in fast_models:
             # We add a vanilla hidden layer that's all
-            x = GlobalAveragePooling1D()(embedding)
-            x = Dense(dense_layer1, activation="relu")(x)
-            x = Dropout(0.5)(x)
-            nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(x)
+            meta_inputs = []
+            nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(embedding)
         elif keras_model_type.lower() in ['gru','cnn2']:
             #### Use this only for Binary-Class classification problems ########
             ####  LSTM with 1D convnet with maxpooling ########
-            x = Conv1D(filters, 
+            x = Conv1D(filters,
                              kernel_size,
                              padding='valid',
                              activation='relu',
@@ -283,7 +277,7 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
                 #nlp_outputs = Dense(class_size, activation='sigmoid')(x)
                 x = Dense(128, activation='relu')(x)
             nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(x)
-        else:
+        elif keras_model_type.lower() in ['cnn']:
             #### Use this only for future Multi-Class classification problems #########
             ####  CNN Model: create a 1D convnet with global maxpooling ########
             x = Conv1D(128, kernel_size, activation='relu')(embedding)
@@ -296,34 +290,28 @@ def perform_preprocessing(train_ds, var_df, cat_vocab_dict, keras_model_type,
             #nlp_outputs = Dense(class_size, activation='softmax')(x)
             nlp_outputs = layers.Dense(num_predicts, activation=output_activation)(x)
 
-    if keras_model_type.lower() in fast_models:
-        print('    All Non-NLP feature preprocessing for %s completed.' %keras_model_type)
-        meta_inputs = []
-        meta_outputs = []
-        return nlp_inputs, meta_inputs, meta_outputs, nlp_outputs
-    else:
-        #### This is only for non-fast but "auto" with latitude and longitude columns ## 
-        if isinstance(meta_outputs, list):
-            ### if meta_outputs is a list, it means there is no int, float or cat variable in this data set
-            print('There is no numeric or cat or int variables in this data set.')
-            if isinstance(nlp_outputs, list):
-                ### if NLP_outputs is a list, it means there is no NLP variable in the data set
-                print('    There is no NLP variable in this data set. Returning')
-            else:
-                print('    Number of predictions = %s from NLP variable fed into model' %(nlp_outputs.shape,))
-                consolidated_outputs = nlp_outputs
+    #### This is only for all "fast" and "auto" with latitude and longitude columns ##
+    if isinstance(meta_outputs, list):
+        ### if meta_outputs is a list, it means there is no int, float or cat variable in this data set
+        print('There is no numeric or cat or int variables in this data set.')
+        if isinstance(nlp_outputs, list):
+            ### if NLP_outputs is a list, it means there is no NLP variable in the data set
+            print('    There is no NLP variable in this data set. Returning')
+            consolidated_outputs = meta_outputs
         else:
-            print('    Shape of output from numeric+integer+cat variables before model training = %s' %(meta_outputs.shape,))
-            if isinstance(nlp_outputs, list):
-                ### if NLP_outputs is a list, it means there is no NLP variable in the data set
-                print('    There is no NLP variable in this data set. Continuing...')
-                #x = layers.concatenate([meta_outputs])
-                consolidated_outputs = meta_outputs
-            else:
-                ### if NLP_outputs is NOT a list, it means there is some NLP variable in the data set
-                print('    Number of predictions = %s from NLP variable fed into model' %(nlp_outputs.shape,))
-                consolidated_outputs = layers.concatenate([nlp_outputs, meta_outputs])
-
-        print('Shape of output from all preprocessing layers before model training = %s' %(consolidated_outputs.shape,))
-        return nlp_inputs, meta_inputs, consolidated_outputs, nlp_outputs
-##########################################################################################
\ No newline at end of file
+            print('    Number of predictions = %s from NLP variable fed into model' %(nlp_outputs.shape,))
+            consolidated_outputs = nlp_outputs
+    else:
+        print('    Shape of output from numeric+integer+cat variables before model training = %s' %(meta_outputs.shape,))
+        if isinstance(nlp_outputs, list):
+            ### if NLP_outputs is a list, it means there is no NLP variable in the data set
+            print('    There is no NLP variable in this data set. Continuing...')
+            #x = layers.concatenate([meta_outputs])
+            consolidated_outputs = meta_outputs
+        else:
+            ### if NLP_outputs is NOT a list, it means there is some NLP variable in the data set
+            print('    Number of predictions = %s from NLP variable fed into model' %(nlp_outputs.shape,))
+            consolidated_outputs = layers.concatenate([nlp_outputs, meta_outputs])
+            print('Shape of output from all preprocessing layers before model training = %s' %(consolidated_outputs.shape,))
+    return nlp_inputs, meta_inputs, consolidated_outputs, nlp_outputs
+##########################################################################################
